from __future__ import absolute_import


try:
    is_stub = False  # type: bool
    """Indicates whether stub is used instead of real `uwsgi` module."""

    ####################################################################

    from uwsgi import *

except ImportError:

    is_stub = True  # type: bool
    """Indicates whether stub is used instead of real `uwsgi` module."""

    ####################################################################

    from .uwsgi_stub import *



from .alarms import AlarmType
from .monitoring_collectors import *
from .monitoring_pushers import *
from .monitoring_metric_types import *
from ..base import OptionsGroup
from ..utils import make_key_val_string, listify


class Monitoring(OptionsGroup):
    """Monitoring facilities.

    * SNMP - http://uwsgi.readthedocs.io/en/latest/SNMP.html

    * Stats - http://uwsgi.readthedocs.io/en/latest/StatsServer.html
        Set of metrics gathered from uWSGI internals.

    * Metrics - http://uwsgi.readthedocs.io/en/latest/Metrics.html
        Basic set of metrics gathered from uWSGI internals + user defined metrics.

    """
    class metric_types(object):
        """Various metric types to represent data of various nature.

        User metrics must inherit from one of those.

        """
        absolute = MetricTypeAbsolute
        alias = MetricTypeAlias
        counter = MetricTypeCounter
        gauge = MetricTypeGauge

    class collectors(object):
        """Metric collection and accumulation means."""

        accumulator = CollectorAccumulator
        adder = CollectorAdder
        avg = CollectorAvg
        file = CollectorFile
        function = CollectorFunction
        multiplier = CollectorMultiplier
        pointer = CollectorPointer
        sum = CollectorSum

    class pushers(object):
        """Means to deliver metrics to various remotes or locals.

        These are available for ``.register_stats_pusher()``.

        """
        carbon = PusherCarbon
        file = PusherFile
        mongo = PusherMongo
        rrdtool = PusherRrdtool
        socket = PusherSocket
        statsd = PusherStatsd
        zabbix = PusherZabbix

    def register_metric(self, metric):
        """

        Officially Registered Metrics:

            * ``worker`` 3 - exports information about workers.
                Example: worker.1.requests **[or 3.1.1]** reports the number of requests served by worker 1.

            * ``plugin`` 4  - namespace for metrics automatically added by plugins.
                Example: plugins.foo.bar

            * ``core`` 5  - namespace for general instance information.

            * ``router`` 6 - namespace for corerouters.
                Example: router.http.active_sessions

            * ``socket`` 7 - namespace for sockets.
                Example: socket.0.listen_queue

            * ``mule`` 8 - namespace for mules.
                Example:  mule.1.signals

            * ``spooler`` 9 - namespace for spoolers.
                Example: spooler.1.signals

            * ``system`` 10 - namespace for system metrics, like loadavg or free memory.

        :param Metric|list[Metric] metric: Metric object.

        """
        for metric in listify(metric):
            self._set('metric', metric, multi=True)

        return self._section

    def set_metrics_params(self, enable=None, store_dir=None, restore=None, no_cores=None):
        """Sets basic Metrics subsystem params.

        uWSGI metrics subsystem allows you to manage "numbers" from your apps.

        When enabled, the subsystem configures a vast amount of metrics
        (like requests per-core, memory usage, etc) but, in addition to this,
        you can configure your own metrics, such as the number of active users or, say,
        hits of a particular URL, as well as the memory consumption of your app or the whole server.

        * http://uwsgi.readthedocs.io/en/latest/Metrics.html
        * SNMP Integration - http://uwsgi.readthedocs.io/en/latest/Metrics.html#snmp-integration

        :param bool enable: Enables the subsystem.

        :param str|unicode store_dir: Directory to store metrics.
            The metrics subsystem can expose all of its metrics in the form
            of text files in a directory. The content of each file is the value
            of the metric (updated in real time).

        :param bool restore: Restore previous metrics from ``store_dir``.
            When you restart a uWSGI instance, all of its metrics are reset.
            Use the option to force the metric subsystem to read-back the values
            from the metric directory before starting to collect values.

        :param bool no_cores: Disable generation of cores-related metrics.

        """
        self._set('enable-metrics', enable, cast=bool)
        self._set('metrics-dir', store_dir)
        self._set('metrics-dir-restore', restore, cast=bool)
        self._set('metrics-no-cores', no_cores, cast=bool)

        return self._section

    def set_metrics_threshold(self, name, value, check_interval=None, reset_to=None, alarm=None, alarm_message=None):
        """Sets metric threshold parameters.

        :param str|unicode name: Metric name.

        :param int value: Threshold value.

        :param int reset_to: Reset value to when threshold is reached.

        :param int check_interval: Threshold check interval in seconds.

        :param str|unicode|AlarmType alarm: Alarm to trigger when threshold is reached.

        :param str|unicode alarm_message: Message to pass to alarm. If not set metrics name is passed.

        """
        if alarm is not None and isinstance(alarm, AlarmType):
            self._section.alarms.register_alarm(alarm)
            alarm = alarm.alias

        value = make_key_val_string(
            locals(),
            aliases={
                'name': 'key',
                'reset_to': 'reset',
                'check_interval': 'rate',
                'alarm_message': 'msg',
            },
        )

        self._set('metric-threshold', value, multi=True)

        return self._section

    def set_stats_params(
            self, address=None, enable_http=None,
            minify=None, no_cores=None, no_metrics=None, push_interval=None):
        """Enables stats server on the specified address.

        * http://uwsgi.readthedocs.io/en/latest/StatsServer.html

        :param str|unicode address: Address/socket to make stats available on.

            Examples:
                * 127.0.0.1:1717
                * /tmp/statsock
                * :5050

        :param bool enable_http: Server stats over HTTP.
            Prefixes stats server json output with http headers.

        :param bool minify: Minify statistics json output.

        :param bool no_cores: Disable generation of cores-related stats.

        :param bool no_metrics: Do not include metrics in stats output.

        :param int push_interval: Set the default frequency of stats pushers in seconds/

        """
        self._set('stats-server', address)
        self._set('stats-http', enable_http, cast=bool)
        self._set('stats-minified', minify, cast=bool)
        self._set('stats-no-cores', no_cores, cast=bool)
        self._set('stats-no-metrics', no_metrics, cast=bool)
        self._set('stats-pusher-default-freq', push_interval)

        return self._section

    def register_stats_pusher(self, pusher):
        """Registers a pusher to be used for pushing statistics to various remotes/locals.

        :param Pusher|list[Pusher] pusher:

        """
        for pusher in listify(pusher):
            self._set('stats-push', pusher, multi=True)

        return self._section

    def enable_snmp(self, address, community_string):
        """Enables SNMP.

        uWSGI server embeds a tiny SNMP server that you can use to integrate
        your web apps with your monitoring infrastructure.

        * http://uwsgi.readthedocs.io/en/latest/SNMP.html

        .. note:: SNMP server is started in the master process after dropping the privileges.
            If you want it to listen on a privileged port, you can either use Capabilities on Linux,
            or use the ``as-root`` option to run the master process as root.

        :param str|unicode address: UDP address to bind to.

            Examples:

                * 192.168.1.1:2222

        :param str|unicode community_string: SNMP instance identifier to address it.

        """
        self._set('snmp', address)
        self._set('snmp-community', community_string)

        return self._section



import sys
from contextlib import contextmanager

import click

from uwsgiconf import VERSION
from uwsgiconf.utils import ConfModule, UwsgiRunner
from uwsgiconf.exceptions import ConfigurationError


@contextmanager
def errorprint():
    """Print out descriptions from ConfigurationError."""
    try:
        yield

    except ConfigurationError as e:
        click.secho('%s' % e, err=True, fg='red')
        sys.exit(1)


@click.group()
@click.version_option(version='.'.join(map(str, VERSION)))
def base():
    """uwsgiconf command line utility.

    Tools to facilitate uWSGI configuration.

    """

arg_conf = click.argument(
    'conf', type=click.Path(exists=True, dir_okay=False), default=ConfModule.default_name)


@base.command()
@arg_conf
def run(conf):
    """Runs uWSGI passing to it using the default or another `uwsgiconf` configuration module.

    """
    with errorprint():
        config = ConfModule(conf)
        config.spawn_uwsgi()


@base.command()
@arg_conf
def compile(conf):
    """Compiles classic uWSGI configuration file using the default
    or given `uwsgiconf` configuration module.

    """
    with errorprint():
        config = ConfModule(conf)
        for conf in config.configurations:
            conf.format(do_print=True)


@base.command()
def probe_plugins():
    """Runs uWSGI to determine what plugins are available and prints them out.

    Generic plugins come first then after blank line follow request plugins.

    """
    plugins = UwsgiRunner().get_plugins()

    for plugin in sorted(plugins.generic):
        click.secho(plugin)

    click.secho('')

    for plugin in sorted(plugins.request):
        click.secho(plugin)


def main():
    """
    CLI entry point
    """
    base(obj={})


if __name__ == '__main__':
    main()



import pytest
import subprocess


@pytest.fixture(scope='session')
def ini():

    def wrapped(section, stamp):
        return section.as_configuration().format(stamp=stamp)

    return wrapped


@pytest.fixture(scope='session')
def assert_lines(ini):

    def wrapped(lines, source, assert_in=True, stamp=False):
        source = ini(source, stamp)

        if not isinstance(lines, list):
            lines = [lines]

        for line in lines:
            if assert_in:
                assert line in source
            else:
                assert line not in source

    return wrapped


@pytest.fixture
def mock_popen(monkeypatch):

    def mock(func_communicate):

        class MockPopen(object):

            def __init__(self, *args, **kwargs):
                pass

            def communicate(self, *args, **kwargs):
                return func_communicate(*args, **kwargs)

        monkeypatch.setattr(subprocess, 'Popen', MockPopen)

    return mock



import os

import pytest

from uwsgiconf.utils import UwsgiRunner, parse_command_plugins_output, ConfModule, get_uwsgi_stub_attrs_diff
from uwsgiconf.exceptions import ConfigurationError, UwsgiconfException


SAMPLE_OUT_PLUGINS_MANY = '''

*** uWSGI loaded generic plugins ***
gevent
syslog
stats_pusher_socket

*** uWSGI loaded request plugins ***
0: python
173: rpc
--- end of plugins list ---

*** Starting uWSGI 2.0.15 (64bit) on [Fri Jul 28 20:05:53 2017] ***
compiled with version: 6.3.0 20170406 on 28 July 2017 20:04:29
'''

SAMPLE_OUT_PLUGINS_EMPTY = '''

*** uWSGI loaded generic plugins ***

*** uWSGI loaded request plugins ***
--- end of plugins list ---

*** Starting uWSGI 2.0.14-debian (64bit) on [Fri Jul 28 20:09:00 2017] ***
compiled with version: 6.3.0 20170221 on 27 February 2017 15:11:38
'''


def test_parser():
    plugins = parse_command_plugins_output(SAMPLE_OUT_PLUGINS_MANY)

    assert len(plugins.generic) == 3
    assert len(plugins.request) == 2
    assert 'rpc' in plugins.request

    plugins = parse_command_plugins_output(SAMPLE_OUT_PLUGINS_EMPTY)
    assert len(plugins.generic) == 0
    assert len(plugins.request) == 0


def test_runner(mock_popen):

    mock_popen(lambda: (SAMPLE_OUT_PLUGINS_MANY, ''))

    runner = UwsgiRunner()
    plugins = runner.get_plugins()
    assert len(plugins.generic) == 3
    assert len(plugins.request) == 2


def test_conf_module_compile():
    fpath = os.path.join(os.path.dirname(__file__), 'confs', 'dummy.py')

    # no attr
    module = ConfModule(fpath)
    module.confs_attr_name = 'faked'
    with pytest.raises(ConfigurationError):
        confs = module.configurations

    # empty
    module = ConfModule(fpath)
    module.confs_attr_name = 'not_conf1'
    with pytest.raises(ConfigurationError):
        confs = module.configurations

    # invalid objects
    module = ConfModule(fpath)
    module.confs_attr_name = 'not_conf2'
    with pytest.raises(ConfigurationError):
        confs = module.configurations

    # invalid objects
    module = ConfModule(fpath)
    assert module.configurations
    assert len(module.configurations) == 1


def test_conf_module_run(monkeypatch):
    executed = []
    monkeypatch.setattr(os, 'execvp', lambda *args: executed.append(True))

    fpath = os.path.join(os.path.dirname(__file__), 'confs', 'dummy.py')

    module = ConfModule(fpath)
    module.spawn_uwsgi()

    assert all(executed)


def test_get_uwsgi_stub_attrs_diff():

    with pytest.raises(UwsgiconfException):
        get_uwsgi_stub_attrs_diff()  # uwsgi is unavailable

    import sys

    sys.modules['uwsgi'] = type('DummyModule', (object,), {'dummy_uwsgi': True})

    from_candidate, from_stub = get_uwsgi_stub_attrs_diff()

    assert from_candidate == ['dummy_uwsgi']
    assert from_stub

    del sys.modules['uwsgi']



import pytest

from uwsgiconf import Section
from uwsgiconf.exceptions import ConfigurationError


def test_subscriptions_basics(assert_lines):

    assert_lines([
        'subscription-mountpoint = 2',
    ], Section().subscriptions.set_server_params(mountpoints_depth=2))

    assert_lines([
        'subscriptions-sign-check',
    ], Section().subscriptions.set_server_verification_params(digest_algo='SHA1'), assert_in=False)

    assert_lines([
        'subscriptions-sign-check = SHA1:/here',
        'subscriptions-sign-skip-uid = 1001',
        'subscriptions-sign-skip-uid = 1002',
    ], Section().subscriptions.set_server_verification_params(
        digest_algo='SHA1', dir_cert='/here', no_check_uid=[1001, 1002]))

    assert_lines([
        'start-unsubscribed = true',
    ], Section().subscriptions.set_client_params(start_unsubscribed=True))

    # Subscribing:

    with pytest.raises(ConfigurationError):  # both key and server omitted
        Section().subscriptions.subscribe(balancing_weight=2)

    assert_lines([
        'key=pythonz.net',
        'server=127.0.0.1:4040',
    ], Section().subscriptions.subscribe('127.0.0.1:4040', 'pythonz.net'))

    # SNI
    assert_lines([
        'socket=0',
        'key=mydomain.it',
        'sni_crt=/foo/bar.crt,sni_key=/foo/bar.key',
    ], Section().subscriptions.subscribe(key='mydomain.it', address=0, sni_cert='/foo/bar.crt', sni_key='/foo/bar.key'))

    # algos
    algo = Section.subscriptions.algorithms.weighted_least_reference_count(2)
    assert_lines([
        'backup=2',
        'algo=wlrc',
    ], Section().subscriptions.subscribe('127.0.0.1:4040', balancing_algo=algo))

    # modifiers
    assert_lines([
        'modifier1=31',
        'modifier2=42',
    ], Section().subscriptions.subscribe('127.0.0.1:4040', modifier=Section.routing.modifiers.message(42)))

    # signing
    assert_lines([
        'sign=SHA1:myssh001',
    ], Section().subscriptions.subscribe('127.0.0.1:4040', signing=('SHA1', 'myssh001')))



from ..base import ParametrizedValue
from ..utils import listify, filter_locals, make_key_val_string


class AlarmType(ParametrizedValue):

    def __init__(self, alias, *args, **kwargs):
        self.alias = alias or ''
        super(AlarmType, self).__init__(*args)


class AlarmCommand(AlarmType):
    """Run a shell command, passing info into its stdin."""

    name = 'cmd'

    def __init__(self, alias, command):
        super(AlarmCommand, self).__init__(alias, command)


class AlarmSignal(AlarmType):
    """Raise an uWSGI signal."""

    name = 'signal'

    def __init__(self, alias, sig_num):
        super(AlarmSignal, self).__init__(alias, sig_num)


class AlarmLog(AlarmType):
    """Print line into log."""

    name = 'log'

    def __init__(self, alias):
        super(AlarmLog, self).__init__(alias)


class AlarmMule(AlarmType):
    """Send info to a mule waiting for messages."""

    name = 'mule'

    def __init__(self, alias, mule_id):
        super(AlarmMule, self).__init__(alias, mule_id)


class AlarmCurl(AlarmType):
    """Send info to a cURL-able URL."""

    name = 'curl'
    plugin = 'alarm_curl'
    args_joiner = ';'

    def __init__(
            self, alias, url, method=None, ssl=None, ssl_insecure=None,
            auth_user=None, auth_pass=None,
            timeout=None, conn_timeout=None,
            mail_from=None, mail_to=None, subject=None):

        opts = make_key_val_string(
            filter_locals(locals(), drop=['alias', 'url']),
            bool_keys=['ssl', 'ssl_insecure'],
            items_separator=self.args_joiner,
        )
        super(AlarmCurl, self).__init__(alias, url, opts)


class AlarmXmpp(AlarmType):
    """Send info via XMPP/jabber."""

    name = 'xmpp'
    plugin = 'alarm_xmpp'
    args_joiner = ';'

    def __init__(self, alias, jid, password, recipients):
        super(AlarmXmpp, self).__init__(alias, jid, password, ','.join(listify(recipients)))



from ..base import OptionsGroup
from ..exceptions import ConfigurationError
from ..utils import make_key_val_string, filter_locals
from .subscriptions_algos import *


class Subscriptions(OptionsGroup):
    """
    This allows some uWSGI instances to announce their presence to subscriptions managing server,
    which in its turn can address those nodes (e.g. delegate request processing to them)
    and automatically remove dead nodes from the pool.

    .. note:: Subscription system in many ways relies on Master Process.

    .. warning:: The subscription system is meant for "trusted" networks.
        All of the nodes in your network can potentially make a total mess with it.

    * http://uwsgi.readthedocs.io/en/latest/SubscriptionServer.html

    """

    class algorithms(object):
        """Balancing algorithms available to use with ``subscribe``."""

        ip_hash = IpHash
        least_reference_count = LeastReferenceCount
        weighted_least_reference_count = WeightedLeastReferenceCount
        weighted_round_robin = WeightedRoundRobin

    def set_server_params(
            self, client_notify_address=None, mountpoints_depth=None, require_vassal=None,
            tolerance=None, tolerance_inactive=None, key_dot_split=None):
        """Sets subscription server related params.

        :param str|unicode client_notify_address: Set the notification socket for subscriptions.
            When you subscribe to a server, you can ask it to "acknowledge" the acceptance of your request.
            pointing address (Unix socket or UDP), on which your instance will bind and
            the subscription server will send acknowledgements to.

        :param int mountpoints_depth: Enable support of mountpoints of certain depth for subscription system.

            * http://uwsgi-docs.readthedocs.io/en/latest/SubscriptionServer.html#mountpoints-uwsgi-2-1

        :param bool require_vassal: Require a vassal field (see ``subscribe``) from each subscription.

        :param int tolerance: Subscription reclaim tolerance (seconds).

        :param int tolerance_inactive: Subscription inactivity tolerance (seconds).

        :param bool key_dot_split: Try to fallback to the next part in (dot based) subscription key.
            Used, for example, in SNI.

        """
        # todo notify-socket (fallback) relation
        self._set('subscription-notify-socket', client_notify_address)
        self._set('subscription-mountpoint', mountpoints_depth)
        self._set('subscription-vassal-required', require_vassal, cast=bool)
        self._set('subscription-tolerance', tolerance)
        self._set('subscription-tolerance-inactive', tolerance_inactive)
        self._set('subscription-dotsplit', key_dot_split, cast=bool)

        return self._section

    def set_server_verification_params(
            self, digest_algo=None, dir_cert=None, tolerance=None, no_check_uid=None,
            dir_credentials=None, pass_unix_credentials=None):
        """Sets peer verification params for subscription server.

        These are for secured subscriptions.

        :param str|unicode digest_algo: Digest algorithm. Example: SHA1

            .. note:: Also requires ``dir_cert`` to be set.

        :param str|unicode dir_cert: Certificate directory.

            .. note:: Also requires ``digest_algo`` to be set.

        :param int tolerance: Maximum tolerance (in seconds) of clock skew for secured subscription system.
            Default: 24h.

        :param str|unicode|int|list[str|unicode|int] no_check_uid: Skip signature check for the specified uids
            when using unix sockets credentials.

        :param str|unicode|list[str|unicode] dir_credentials: Directories to search for subscriptions
            key credentials.

        :param bool pass_unix_credentials: Enable management of SCM_CREDENTIALS in subscriptions UNIX sockets.

        """
        if digest_algo and dir_cert:
            self._set('subscriptions-sign-check', '%s:%s' % (digest_algo, dir_cert))

        self._set('subscriptions-sign-check-tolerance', tolerance)
        self._set('subscriptions-sign-skip-uid', no_check_uid, multi=True)
        self._set('subscriptions-credentials-check', dir_credentials, multi=True)
        self._set('subscriptions-use-credentials', pass_unix_credentials, cast=bool)

        return self._section

    def set_client_params(
            self, start_unsubscribed=None, clear_on_exit=None, unsubscribe_on_reload=None,
            announce_interval=None):
        """Sets subscribers related params.

        :param bool start_unsubscribed: Configure subscriptions but do not send them.
            .. note:: Useful with master FIFO.

        :param bool clear_on_exit: Force clear instead of unsubscribe during shutdown.

        :param bool unsubscribe_on_reload: Force unsubscribe request even during graceful reload.

        :param int announce_interval: Send subscription announce at the specified interval. Default: 10 master cycles.

        """
        self._set('start-unsubscribed', start_unsubscribed, cast=bool)
        self._set('subscription-clear-on-shutdown', clear_on_exit, cast=bool)
        self._set('unsubscribe-on-graceful-reload', unsubscribe_on_reload, cast=bool)
        self._set('subscribe-freq', announce_interval)

        return self._section

    def subscribe(
            self, server=None, key=None, address=None, address_vassal=None,
            balancing_weight=None, balancing_algo=None, modifier=None, signing=None, check_file=None, protocol=None,
            sni_cert=None, sni_key=None, sni_client_ca=None):
        """Registers a subscription intent.

        :param str|unicode server: Subscription server address (UDP or UNIX socket).

            Examples:
                * 127.0.0.1:7171

        :param str|unicode key: Key to subscribe. Generally the domain name (+ optional '/< mountpoint>').
            Examples:
                * mydomain.it/foo
                * mydomain.it/foo/bar (requires ``mountpoints_depth=2``)
                * mydomain.it
                * ubuntu64.local:9090

        :param str|unicode|int address: Address to subscribe (the value for the key)
            or zero-based internal socket number (integer).

        :param str|unicode address: Vassal node address.

        :param int balancing_weight: Load balancing value. Default: 1.

        :param balancing_algo: Load balancing algorithm to use. See ``balancing_algorithms``
            .. note:: Since 2.1

        :param Modifier modifier: Routing modifier object. See ``.routing.modifiers``

        :param list|tuple signing:  Signing basics, expects two elements list/tuple:
            (signing_algorithm, key).

            Examples:
                * SHA1:idlessh001

        :param str|unicode check_file: If this file exists the subscription packet is sent,
            otherwise it is skipped.

        :param str|unicode protocol: the protocol to use, by default it is ``uwsgi``.
            See ``.networking.socket_types``.

            .. note:: Since 2.1

        :param str|unicode sni_cert: Certificate file to use for SNI proxy management.
            * http://uwsgi.readthedocs.io/en/latest/SNI.html#subscription-system-and-sni

        :param str|unicode sni_key: sni_key Key file to use for SNI proxy management.
            * http://uwsgi.readthedocs.io/en/latest/SNI.html#subscription-system-and-sni

        :param str|unicode sni_client_ca: Ca file to use for SNI proxy management.
            * http://uwsgi.readthedocs.io/en/latest/SNI.html#subscription-system-and-sni

        """
        # todo params: inactive (inactive slot activation)

        if not any((server, key)):
            raise ConfigurationError('Subscription requires `server` or `key` to be set.')

        address_key = 'addr'
        if isinstance(address, int):
            address_key = 'socket'

        if balancing_algo:
            backup = getattr(balancing_algo, 'backup_level', None)

        if signing:
            signing = ':'.join(signing)

        if modifier:
            modifier1 = modifier
            if modifier.submod:
                modifier2 = modifier.submod

        rule = make_key_val_string(
            filter_locals(locals(), drop=['address_key', 'modifier']),
            aliases={
                'address': address_key,
                'address_vassal': 'vassal',
                'signing': 'sign',
                'check_file': 'check',
                'balancing_weight': 'weight',
                'balancing_algo': 'algo',
                'protocol': 'proto',
                'sni_cert': 'sni_crt',
                'sni_client_ca': 'sni_ca',
            },
        )

        self._set('subscribe2', rule)

        return self._section



from ..base import ParametrizedValue
from ..utils import make_key_val_string, filter_locals
from ..exceptions import ConfigurationError


class Metric(ParametrizedValue):

    type_str = None
    type_id = None
    name_separator = ','

    def __init__(
            self, name, oid=None, alias_for=None, collector=None,
            initial_value=None, collect_interval=None, reset_after_push=None):
        """

        :param str|unicode name: Metric name.

            .. note:: Only numbers, letters, underscores, dashes and dots.

        :param str|unicode alias_for: If set metric will be a simple alias for the specified one.

        :param str|unicode oid: Metric OID.

            Required for SNMP.

            * http://uwsgi-docs.readthedocs.io/en/latest/Metrics.html#oid-assigment-for-plugins

        :param Collector collector: Collector to be used. If not set it is considered that the value must
            be updated manually from applications using the metrics API.

            * http://uwsgi-docs.readthedocs.io/en/latest/Metrics.html#api

        :param int initial_value: Set the metric to a specific value on startup.

        :param int collect_interval: How ofter the metric should be gathered. In seconds.

        :param bool reset_after_push: Reset the metric to zero (or the configured initial_value)
            after it's been pushed.

        """
        self.name = name

        if alias_for:
            # Set any type to ``alias``.
            self.type_str = MetricTypeAlias.type_str
            self.type_id = MetricTypeAlias.type_id

            if isinstance(alias_for, Metric):
                alias_for = alias_for.name

        if self.type_id == MetricTypeAlias.type_id and not alias_for:
            raise ConfigurationError('Parameter ``alias_for`` is required for ``MetricTypeAlias``.')

        type_str = self.type_str

        value = make_key_val_string(
            filter_locals(locals(), drop=['name']),
            aliases={'collect_interval': 'freq', 'type_str': 'type', 'alias_for': 'alias'}
        )

        super(Metric, self).__init__(value)

    def __str__(self):
        return 'name=%s' % super(Metric, self).__str__()


class MetricTypeCounter(Metric):
    """A generally-growing up number.

        Example:
            * number of requests

    """
    type_str = 'counter'
    type_id = 0


class MetricTypeGauge(Metric):
    """A number that can increase or decrease dynamically.

        Example:
            * memory used by a worker
            * CPU load

    """
    type_str = 'gauge'
    type_id = 1


class MetricTypeAbsolute(Metric):
    """An absolute number.

        Example:
            * memory of the whole server
            * size of the hard disk.

    """
    type_str = 'absolute'
    type_id = 2


class MetricTypeAlias(Metric):
    """This is a virtual metric pointing to another one .

    You can use it to give different names to already existing metrics.

    """
    type_str = 'alias'
    type_id = 3



from __future__ import unicode_literals, absolute_import
import os
import sys
from contextlib import contextmanager
from collections import namedtuple
from importlib import import_module

try:  # pragma: nocover
    from StringIO import StringIO
except ImportError:  # pragma: nocover
    from io import StringIO

from .exceptions import ConfigurationError


EmbeddedPlugins = namedtuple('EmbeddedPlugins', ['generic', 'request'])


@contextmanager
def output_capturing():
    """Temporarily captures/redirects stdout."""
    out = sys.stdout

    sys.stdout = StringIO()

    try:
        yield

    finally:
        sys.stdout = out


class ConfModule(object):
    """Represents a uwsgiconf configuration module.

    Allows reading configurations from .py files.

    """
    default_name = 'uwsgicfg.py'
    confs_attr_name = 'configuration'

    def __init__(self, fpath):
        """Module filepath.

        :param fpath:
        """
        fpath = os.path.abspath(fpath)
        self.fpath = fpath
        self._confs = None

    def spawn_uwsgi(self):
        """Spawns uWSGI process wich will use current configuration module.

        .. note:: uWSGI process will replace ``uwsgiconf`` process.

        """
        UwsgiRunner().spawn(self.fpath)

    @property
    def configurations(self):
        """Configuration( section(s) from uwsgiconf module."""

        if self._confs is not None:
            return self._confs

        from uwsgiconf.config import Section, Configuration

        fpath = self.fpath

        with output_capturing():
            module = self.load(fpath)

            attr = getattr(module, self.confs_attr_name, None)

            if attr is None:
                raise ConfigurationError(
                    "'configuration' attribute not found [ %s ]" % fpath)

            if callable(attr):
                attr = attr()

        if not attr:
            raise ConfigurationError(
                "'configuration' attribute is empty [ %s ]" % fpath)

        if not isinstance(attr, (list, tuple)):
            attr = [attr]

        confs = []

        for conf_candidate in attr:
            if not isinstance(conf_candidate, (Section, Configuration)):
                continue

            if isinstance(conf_candidate, Section):
                conf_candidate = conf_candidate.as_configuration()

            confs.append(conf_candidate)

        if not confs:
            raise ConfigurationError(
                "'configuration' attribute must hold either 'Section' nor 'Configuration' objects [ %s ]" % fpath)

        self._confs = confs

        return confs

    @classmethod
    def load(cls, fpath):
        """Loads a module and returns its object.

        :param str|unicode fpath:
        :rtype: module
        """
        module_name = os.path.splitext(os.path.basename(fpath))[0]

        sys.path.insert(0, os.path.dirname(fpath))
        try:
            module = import_module(module_name)

        finally:
            sys.path = sys.path[1:]

        return module


def listify(src):
    """Make a list with source object if not already a list.

    :param src:
    :rtype: list

    """
    if not isinstance(src, list):
        src = [src]

    return src


def filter_locals(locals_dict, drop=None):
    """Filters a dictionary produced by locals().

    :param dict locals_dict:

    :param list drop: Keys to drop from dict.

    :rtype: dict
    """
    drop = drop or []
    drop.extend([
        'self',
        '__class__',  # py3
    ])
    locals_dict = {k: v for k, v in locals_dict.items() if k not in drop}
    return locals_dict


def make_key_val_string(
        locals_dict, keys=None, aliases=None, bool_keys=None, list_keys=None,
        items_separator=','):
    """Flattens the given dictionary into key-value string.

    :param dict locals_dict: Dictionary produced by locals().

    :param list keys: Relevant keys from dictionary.
        If not defined - all keys are relevant.
        If defined keys will flattened into string using given order.

    :param dict aliases: Mapping key names from locals_dict into names
        they should be replaced with.

    :param list bool_keys: Keys to consider their values bool.

    :param list list_keys: Keys expecting lists.

    :param str|unicode items_separator: String to use as items (chunks) separator.

    :rtype: str|unicode
    """
    value_chunks = []

    keys = keys or sorted(filter_locals(locals_dict).keys())
    aliases = aliases or {}
    bool_keys = bool_keys or []
    list_keys = list_keys or []

    for key in keys:
        val = locals_dict[key]

        if val is not None:

            if key in bool_keys:
                val = 1

            elif key in list_keys:

                val = ';'.join(listify(val))

            value_chunks.append('%s=%s' % (aliases.get(key, key), val))

    return items_separator.join(value_chunks).strip()


class UwsgiRunner(object):
    """Exposes methods to run uWSGI."""

    def __init__(self, binary_path=None):
        self.binary_uwsgi = binary_path or 'uwsgi'
        self.binary_python = self.prepare_env()

    def get_output(self, command_args):
        """Runs a command and returns its output (stdout + stderr).

        :param str|unicode|list[str|unicode] command_args:

        :rtype: str|unicode

        """
        from subprocess import Popen, STDOUT, PIPE

        command = [self.binary_uwsgi]
        command.extend(listify(command_args))

        process = Popen(command, stdout=PIPE, stderr=STDOUT)
        out, _ = process.communicate()

        return out

    def get_plugins(self):
        """Returns ``EmbeddedPlugins`` object with

        :rtype EmbeddedPlugins:
        """
        out = self.get_output('--plugin-list')
        return parse_command_plugins_output(out)

    @classmethod
    def prepare_env(cls):
        """Prepares current environment and returns Python binary name.

        This adds some virtualenv friendliness so that we try use uwsgi from it.

        :rtype: str|unicode
        """
        python_binary = sys.executable
        basepath = os.path.dirname(python_binary)

        os.environ['PATH'] = basepath + os.pathsep + os.environ['PATH']

        return os.path.basename(python_binary)

    def spawn(self, module_path):
        """Spawns uWSGI using the given configuration module.

        .. note:: uWSGI process will replace ``uwsgiconf`` process.

        :param str|unicode module_path:

        """
        os.execvp('uwsgi', ['uwsgi', '--ini', 'exec://%s %s' % (self.binary_python, module_path)])


def parse_command_plugins_output(out):
    """Parses ``plugin-list`` command output from uWSGI
    and returns object containing lists of embedded plugin names.

    :param str|unicode out:

    :rtype EmbeddedPlugins:

    """
    out = out.split('--- end of plugins list ---')[0]
    out = out.partition('plugins ***')[2]
    out = out.splitlines()

    current_slot = 0

    plugins = EmbeddedPlugins([], [])

    for line in out:
        line = line.strip()

        if not line:
            continue

        if line.startswith('***'):
            current_slot += 1
            continue

        if current_slot is not None:
            plugins[current_slot].append(line)

    plugins = plugins._replace(request=[plugin.partition(' ')[2] for plugin in plugins.request])

    return plugins


def get_uwsgi_stub_attrs_diff():
    """Returns attributes difference two elements tuple between
    real uwsgi module and its stub.

    Might be of use while describing in stub new uwsgi functions.

    :return: (uwsgi_only_attrs, stub_only_attrs)

    :rtype: tuple
    """

    try:
        import uwsgi

    except ImportError:
        from uwsgiconf.exceptions import UwsgiconfException

        raise UwsgiconfException(
            '`uwsgi` module is unavailable. Calling `get_attrs_diff` in such environment makes no sense.')

    from . import uwsgi_stub

    def get_attrs(src):
        return set(attr for attr in dir(src) if not attr.startswith('_'))

    attrs_uwsgi = get_attrs(uwsgi)
    attrs_stub = get_attrs(uwsgi_stub)

    from_uwsgi = sorted(attrs_uwsgi.difference(attrs_stub))
    from_stub = sorted(attrs_stub.difference(attrs_uwsgi))

    return from_uwsgi, from_stub



from django.conf import settings
from django.core.management.base import BaseCommand

import os
import sys

from .....presets.nice import PythonSection


def get_section():

    # Expect classic project layout for now.

    DIR_WORKING = os.getcwd()
    PROJECT_NAME = os.path.basename(DIR_WORKING)

    os.environ.setdefault('DJANGO_SETTINGS_MODULE', '%s.settings' % PROJECT_NAME)

    section = PythonSection(
        wsgi_module='%s.wsgi' % PROJECT_NAME,

    ).python.set_basic_params(
        python_home=sys.prefix, search_path=DIR_WORKING

    ).main_process.set_owner_params(
        os.getuid(), os.getegid()

    ).main_process.set_naming_params(
        prefix='[uWSGI %s] ' % PROJECT_NAME
    )

    if settings.DEBUG:
        section.python.set_autoreload_params(2)

        def add_static(url, path):

            if url and path:
                section.statics.register_static_map(url, path)

        # todo Gather static or find a workaround.
        add_static(settings.STATIC_URL, settings.STATIC_ROOT)
        add_static(settings.MEDIA_URL, settings.MEDIA_ROOT)

    section.networking.register_socket(
        '127.0.0.1:8000', type=section.networking.socket_types.HTTP)

    return section


class Command(BaseCommand):

    help = 'Runs uWSGI server'

    def handle(self, *args, **options):

        # todo Support automatic loading from project/uwsgicfg.py
        # todo Support loading from given .py config.
        # todo Support loading from default .py config as in ``get_section``.

        section = get_section()
        filepath = section.as_configuration().tofile()

        os.execvp('uwsgi', ['uwsgi', '--ini=%s' % filepath])



from uwsgiconf.presets.nice import Section, PythonSection


def test_nice_section(assert_lines):

    assert_lines([
        'workers = %k',

    ], Section())

    assert_lines([
        'workers = 13',
        'touch-reload', 'test_nice.py',

    ], Section(workers=13, touch_reload=__file__))


def test_nice_python(assert_lines):

    assert_lines([
        'plugin = python',
        'pyhome = /home/idle/venv/\npythonpath = /home/idle/apps/',
        'wsgi = somepackage.module',

    ], PythonSection(
        params_python=dict(
            # We'll run our app using virtualenv.
            python_home='/home/idle/venv/',
            search_path='/home/idle/apps/',
        ),
        wsgi_module='somepackage.module',
        embedded_plugins=None
    ))

    # Embedded plugins = True
    assert_lines('plugin = python', PythonSection(wsgi_module='somepackage.module'), assert_in=False)



from ..exceptions import ConfigurationError


RANGE_255 = list(range(256))


class Modifier(object):

    _submod = None
    subcodes = None

    code = None
    submod = property()

    def __init__(self, submod=None):
        self.submod = submod

    @classmethod
    def _get_code_str(cls, src):
        """
        :param src:
        :rtype: str|unicode
        """
        return str(src) if src is not None else ''

    def __str__(self):
        return self._get_code_str(self.code)

    @submod.getter
    def submod(self):
        """Modifier 2 value.

        :rtype: str|unicode
        """
        return self._get_code_str(self._submod)

    @submod.setter
    def submod(self, val):
        allowed = self.subcodes

        if val is not None and (allowed is not None and val not in allowed):
            raise ConfigurationError(
                'Unsupported submod (modifier2) value `%s` for %s (modifier1) `%s`' %
                (val, self.__class__.__name__, self.code))

        self._submod = val


class ModifierWsgi(Modifier):
    """Standard WSGI request followed by the HTTP request body."""
    subcodes = [0]
    code = 0


class ModifierPsgi(Modifier):
    """Standard PSGI request followed by the HTTP request body."""
    subcodes = [0]
    code = 5


class ModifierLua(Modifier):
    """Standard LUA/WSAPI request followed by the HTTP request body."""
    subcodes = [0]
    code = 6


class ModifierRack(Modifier):
    """Standard RACK request followed by the HTTP request body."""
    subcodes = [0]
    code = 7


class ModifierJvm(Modifier):
    """Standard JVM request for The JWSGI interface and
    The Clojure/Ring JVM request handler followed by the HTTP request body.

    """
    SUB_RING = 1
    """Use Clojure/Ring JVM request handler."""

    subcodes = [0, SUB_RING]
    code = 8


class ModifierCgi(Modifier):
    """Standard Running CGI scripts on uWSGI request followed by the HTTP request body."""
    subcodes = [0]
    code = 9


class ModifierManage(Modifier):
    """Management interface request: setup flag specified by ``modifier2``.

    For a list of management flag look at ``ManagementFlag``.

    """
    subcodes = RANGE_255
    code = 10


class ModifierGccgo(Modifier):
    code = 11


class ModifierPhp(Modifier):
    """Standard Running PHP scripts in uWSGI request followed by the HTTP request body."""
    subcodes = [0]
    code = 14


class ModifierMono(Modifier):
    """Standard The Mono ASP.NET plugin request followed by the HTTP request body."""
    subcodes = [0]
    code = 15


class ModifierSpooler(Modifier):
    """The uWSGI Spooler request, the block vars is converted
    to a dictionary/hash/table and passed to the spooler callable.

    """
    subcodes = RANGE_255
    code = 17


class ModifierSymcall(Modifier):
    """Direct call to C-like symbols."""
    subcodes = RANGE_255
    code = 18


class ModifierSsi(Modifier):
    code = 19


class ModifierEval(Modifier):
    """Raw Code evaluation. The interpreter is chosen by the ``modifier2``.

    ..note:: It does not return a valid uwsgi response, but a raw string (that may be an HTTP response).

    """
    SUB_PYTHON = 0
    SUB_PERL = 5

    subcodes = [SUB_PYTHON, SUB_PERL]
    code = 22


class ModifierXslt(Modifier):
    """Invoke the The XSLT plugin."""
    subcodes = RANGE_255
    code = 23


class ModifierV8(Modifier):
    """Invoke the uWSGI V8 support."""
    subcodes = RANGE_255
    code = 24


class ModifierGridfs(Modifier):
    """Invoke the The GridFS plugin."""
    subcodes = RANGE_255
    code = 25


class ModifierFastfunc(Modifier):
    """Call the FastFuncs specified by the ``modifier2`` field."""
    subcodes = RANGE_255
    code = 26


class ModifierGlusterfs(Modifier):
    """Invoke the The GlusterFS plugin."""
    subcodes = RANGE_255
    code = 27


class ModifierRados(Modifier):
    """Invoke the The RADOS plugin."""
    subcodes = RANGE_255
    code = 28


class ModifierManagePathInfo(Modifier):
    """Standard WSGI request followed by the HTTP request body.

    The ``PATH_INFO`` is automatically modified, removing the ``SCRIPT_NAME`` from it.

    """
    code = 30


class ModifierMessage(Modifier):
    """Generic message passing (reserved)."""
    subcodes = RANGE_255
    code = 31


class ModifierMessageArray(Modifier):
    """Array of char passing (reserved)."""
    subcodes = RANGE_255
    code = 32


class ModifierMessageMarshal(Modifier):
    """Marshalled/serialzed object passing (reserved)."""
    subcodes = RANGE_255
    code = 33


class ModifierWebdav(Modifier):
    code = 35


class ModifierSnmp(Modifier):
    """Identify a SNMP request/response (mainly via UDP)."""
    code = 48


class ModifierRaw(Modifier):
    """Corresponds to the ``HTTP`` string and signals that
    this is a raw HTTP response.

    """
    code = 72


class ModifierMulticastAnnounce(Modifier):
    """Announce message."""
    code = 73


class ModifierMulticast(Modifier):
    """Array of chars; a custom multicast message managed by uwsgi."""
    subcodes = [0]
    code = 74


class ModifierClusterNode(Modifier):
    """Add/remove/enable/disable node from a cluster.

    Add action requires a dict of at least 3 keys:

        * hostname
        * address
        * workers

    """
    SUB_ADD = 0
    SUB_REMOVE = 1
    SUB_ENABLE = 2
    SUB_DISABLE = 3

    subcodes = [
        SUB_ADD,
        SUB_REMOVE,
        SUB_ENABLE,
        SUB_DISABLE,
    ]
    code = 95


class ModifierRemoteLogging(Modifier):
    """Remote logging (clustering/multicast/unicast)."""
    subcodes = [0]
    code = 96


class ModifierReload(Modifier):
    """Graceful reload request."""

    SUB_REQUEST = 0
    SUB_CONFIRMATION = 1

    subcodes = [SUB_REQUEST, SUB_CONFIRMATION]
    code = 98


class ModifierReloadBrutal(ModifierReload):
    """Brutal reload request."""
    code = 97


class ModifierConfigFromNode(Modifier):
    """Request configuration data from a uwsgi node (even via multicast)."""
    subcodes = [0, 1]
    code = 99


class ModifierPing(Modifier):
    """PING-PONG. Useful for cluster health check."""

    SUB_PING = 0
    """Request."""

    SUB_PONG = 1
    """Response."""

    subcodes = [SUB_PING, SUB_PONG]
    code = 100


class ModifierEcho(Modifier):
    """ECHO service."""
    subcodes = [0]
    code = 101


class ModifierLegionMsg(Modifier):
    """Legion msg (UDP, the body is encrypted)."""
    subcodes = RANGE_255
    code = 109


class ModifierSignal(Modifier):
    """uwsgi_signal framework (payload is optional).

    .. note:: ``modifier2`` is the signal num.

    """
    subcodes = RANGE_255
    code = 110


class ModifierCache(Modifier):
    """Cache operations."""

    SUB_GET = 0
    """Simple cache get for values not bigger than 64k."""

    SUB_SET = 1
    """Simple cache set for values not bigger than 64k."""

    SUB_DELETE = 2
    """Simple cache del."""

    SUB_DICT_BASED = 3
    """Simple dict based get command."""

    SUB_STREAM = 5
    """Get and stream."""

    SUB_DUMP = 6
    """Dump the whole cache."""

    SUB_MAGIC = 17
    """Magic interface for plugins remote access."""

    subcodes = [
        SUB_GET,
        SUB_SET,
        SUB_DELETE,
        SUB_DICT_BASED,
        SUB_STREAM,
        SUB_DUMP,
        SUB_MAGIC,
    ]
    code = 111


class ModifierCorerouterSignal(Modifier):
    """Special modifier for signaling corerouters about special conditions."""
    code = 123


class ModifierRpc(Modifier):
    """RPC. The packet is an uwsgi array where

        * the first item - the name of the function
        * the following - the args

    """
    SUB_DEFAULT = 0
    """Return uwsgi header + rpc response."""

    SUB_RAW = 1
    """Return raw rpc response, uwsgi header included, if available."""

    SUB_USE_PATH_INFO = 2
    """Split PATH_INFO to get func name and args and return as HTTP response 
    with content_type as application/binary or Accept request header (if different from *).
    
    """

    SUB_XMLRPC = 3
    """Set xmlrpc wrapper (requires libxml2)."""

    SUB_JSONRPC = 4
    """Set jsonrpc wrapper (requires libjansson)."""

    SUB_DICT = 5
    """Used in uwsgi response to signal the response is a uwsgi dictionary 
    followed by the body (the dictionary must contains a CONTENT_LENGTH key).
    
    """

    subcodes = [
        SUB_DEFAULT,
        SUB_RAW,
        SUB_USE_PATH_INFO,
        SUB_XMLRPC,
        SUB_JSONRPC,
        SUB_DICT,
    ]
    code = 173


class ModifierPersistentClose(Modifier):
    """Close mark for persistent connections."""
    subcodes = [0]
    code = 200


class ModifierSubscription(Modifier):
    """Subscription packet. See ``subscriptions``."""
    subcodes = [0]
    code = 224


class ModifierExample(Modifier):
    """Modifier used in dummy example plugin."""
    code = 250


class ModifierResponse(Modifier):
    """Generic response. Request dependent.

    Example: a spooler response set 0 for a failed spool or 1 for a successful one.

    """
    subcodes = RANGE_255
    code = 255




SPOOL_IGNORE = 0
"""Spooler function result.

Ignore this task, if multiple languages are loaded in the instance all 
of them will fight for managing the task. 
This return values allows you to skip a task in specific languages.

* http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html#setting-the-spooler-function-callable

"""

SPOOL_OK = -2
"""Spooler function result.

The task has been completed, the spool file will be removed.

* http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html#setting-the-spooler-function-callable

"""

SPOOL_RETRY = -1
"""Spooler function result.

Something is temporarily wrong, the task will be retried at the next spooler iteration.

* http://uwsgi-docs.readthedocs.io/en/latest/Spooler.html#setting-the-spooler-function-callable

"""

SymbolsImporter = None  # type: type
"""SymbolsImporter type."""

SymbolsZipImporter = None  # type: type
"""SymbolsZipImporter type."""

ZipImporter = None  # type: type
"""ZipImporter type."""

applications = None  # type: dict
"""Applications dictionary mapping mountpoints to application callables.

.. note:: Can be ``None``.

* http://uwsgi.readthedocs.io/en/latest/Python.html#application-dictionary

"""

buffer_size = None  # type: int
"""The current configured buffer size in bytes."""

cores = None  # type
"""Detected number of processor cores."""

env = None  # type: dict
"""Request evironment dictionary."""

has_threads = None  # type: bool
"""Flag indicating whether thread support is enabled."""

hostname = None  # type: str
"""Current host name."""

magic_table = None  # type: dict
"""Current mapping of configuration file "magic" variables.

* http://uwsgi.readthedocs.io/en/latest/Configuration.html#magic-variables

"""

numproc = None  # type: int
"""Number of workers (processes) currently running."""

opt = None  # type: dict
"""The current configuration options, including any custom placeholders."""

started_on = 0  # type: int
"""uWSGI's startup Unix timestamp."""

start_response = None  # type: object
"""Callable spitting UWSGI response."""

unbit = None  # type: bool
"""Unbit internal flag."""

version = None  # type: bytes
"""The uWSGI version string.

.. warning:: Bytes are returned for Python 3.

:rtype: bytes|str
"""

version_info = None  # type: tuple
"""Five-elements version number tuple.

:rtype: tuple[int, int, int, int, bytes|str]
"""


def accepting():
    """Called to notify the master that the worker is accepting requests,
    this is required for ``touch_chain_reload`` to work.

    :rtype: None
    """


def add_cron(signal, minute, hour, day, month, weekday):
    """Adds cron. The interface to the uWSGI signal cron facility. The syntax is

    .. note:: The last 5 arguments work similarly to a standard crontab,
    but instead of "*", use -1, and instead of "/2", "/3", etc. use -2 and -3, etc.

    :param int signal: Signal to raise.

    :param int minute: Minute 0-59. Defaults to `each`.

    :param int hour: Hour 0-23. Defaults to `each`.

    :param int day: Day of the month number 1-31. Defaults to `each`.

    :param int month: Month number 1-12. Defaults to `each`.

    :param int weekday: Day of a the week number. Defaults to `each`.
        0 - Sunday  1 - Monday  2 - Tuesday  3 - Wednesday
        4 - Thursday  5 - Friday  6 - Saturday

    :rtype: bool

    :raises ValueError: If unable to add cron rule.
    """


def add_file_monitor(signal, filename):
    """Maps a specific file/directory modification event to a signal.

    :param int signal: Signal to raise.

    :param str|unicode filename: File or a directory to watch for its modification.

    :rtype: None

    :raises ValueError: If unable to register monitor.
    """


def add_ms_timer(signal, period):
    """Add a microsecond resolution timer.

    :param int signal: Signal to raise.

    :param int period: The interval (microseconds) at which to raise the signal.

    :rtype: None

    :raises ValueError: If unable to add timer.
    """


def add_rb_timer(signal, period, repeat=0):
    """Add a red-black timer.

    :param int signal: Signal to raise.

    :param int period: The interval (seconds) at which the signal is raised.

    :param int repeat: How many times to repeat. Default: 0 - infinitely.

    :rtype: None

    :raises ValueError: If unable to add timer.
    """


def add_timer(signal, period):
    """Add timer.

    :param int signal: Signal to raise.

    :param int period: The interval (seconds) at which to raise the signal.

    :rtype: None

    :raises ValueError: If unable to add timer.
    """


def add_var(name, value):
    """Registers custom request variable.

    Can be used for better integration with the internal routing subsystem.

    :param str|unicode name:

    :param str|unicode value:

    :rtype: bool

    :raises ValueError: If buffer size is not enough.
    """


def alarm(name, message):
    """Issues the given alarm with the given message.

    :param str|unicode name:

    :param str|unicode message: Message to pass to alarm.

    :rtype: None
    """


def async_connect(socket):
    """Issues socket connection. And returns a file descriptor.

    * http://uwsgi.readthedocs.io/en/latest/Async.html

    :param str|unicode socket:

    :rtype: int
    """


def async_sleep(seconds):
    """Suspends handling the current request and passes control to the next async core.

    * http://uwsgi.readthedocs.io/en/latest/Async.html

    :param seconds: Sleep time, in seconds.

    :rtype None:
    """


def cache_clear(cache):
    """Clears cache with the given name.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: None
    """


def cache_dec(key, value=1, expires=None, cache=None):
    """Decrements the specified key value by the specified value.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.9.html#math-for-cache

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds).

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_del(key, cache=None):
    """Deletes the given cached key from the cache.

    :param str|unicode key: The cache key to delete.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: None
    """


def cache_div(key, value=2, expires=None, cache=None):
    """Divides the specified key value by the specified value.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.9.html#math-for-cache

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds).

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_exists(key, cache=None):
    """Checks whether there is a value in the cache associated with the given key.

    :param str|unicode key: The cache key to check.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_get(key, cache=None):
    """Gets a value from the cache.

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode key: The cache key to get value for.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bytes|str
    """


def cache_inc(key, value=1, expires=None, cache=None):
    """Increments the specified key value by the specified value.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.9.html#math-for-cache

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds).

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_keys(cache=None):
    """Returns a list of keys available in cache.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: list

    :raises ValueError: If cache is unavailable.
    """


def cache_mul(key, value=2, expires=None, cache=None):
    """Multiplies the specified key value by the specified value.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.9.html#math-for-cache

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds).

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_num(key, cache=None):
    """Gets the 64bit number from the specified item.

    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.9.html#math-for-cache

    :param str|unicode key: The cache key to get value for.

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: int|long
    """


def cache_set(key, value, expires=None, cache=None):
    """Sets the specified key value.

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds). Expire timeout (seconds)

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def cache_update(key, value, expires=None, cache=None):
    """Updates the specified key value.

    :param str|unicode key:

    :param int value:

    :param int expires: Expire timeout (seconds). Expire timeout (seconds)

    :param str|unicode cache: Cache name with optional address (if @-syntax is used).

    :rtype: bool
    """


def call(func_name, *args):
    """Performs an [RPC] function call with the given arguments.

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode func_name: Function name to call
        with optional address (if @-syntax is used).

    :param args:

    :rtype: bytes|str
    """


def chunked_read(timeout):
    """Reads chunked input.

    .. warning:: Bytes are returned for Python 3.

    * http://uwsgi.readthedocs.io/en/latest/Chunked.html
    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.13.html#chunked-input-api

    :param int timeout: Wait timeout (seconds).

    :rtype: bytes|str

    :raises IOError: If unable to receive chunked part.
    """


def chunked_read_nb():
    """Reads chunked input without blocking.

    .. warning:: Bytes are returned for Python 3.

    * http://uwsgi.readthedocs.io/en/latest/Chunked.html
    * http://uwsgi.readthedocs.io/en/latest/Changelog-1.9.13.html#chunked-input-api

    :rtype: bytes|str

    :raises IOError: If unable to receive chunked part.
    """


def cl():
    """Returns current post content length.

    :rtype: int|long
    """


def close(fd):
    """Closes the given file descriptor.

    :param int fd: File descriptor.

    :rtype: None
    """


def connect(socket, timeout=0):
    """Connects to the socket.

    :param str|unicode socket: Socket name.

    :param int timeout: Timeout (seconds).

    :rtype: int
    """


def connection_fd():
    """Returns current request file descriptor.

    :rtype: int
    """


def disconnect():
    """Drops current connection.

    :rtype: None
    """


def embedded_data(symbol_name):
    """Reads a symbol from the uWSGI binary image.

    * http://uwsgi.readthedocs.io/en/latest/Embed.html

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode symbol_name: The symbol name to extract.

    :rtype: bytes|str

    :raises ValueError: If symbol is unavailable.
    """


def extract(fname):
    """Extracts file contents.

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode fname:

    :rtype: bytes|str
    """


def farm_get_msg():
    """Reads a mule farm message.

     * http://uwsgi.readthedocs.io/en/latest/Embed.html

     .. warning:: Bytes are returned for Python 3.

     :rtype: bytes|str|None

     :raises ValueError: If not in a mule
     """


def farm_msg(farm, message):
    """Sends a message to the given farm.

    :param str|unicode farm: Farm name to send message to.

    :param str|unicode message:

    :rtype: None
    """


def get_logvar(name):
    """Return user-defined log variable contents.

    * http://uwsgi.readthedocs.io/en/latest/LogFormat.html#user-defined-logvars

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode name:

    :rtype: bytes|str
    """


def green_schedule():
    """Switches to another green thread.

    .. note:: Alias for ``suspend``.

    * http://uwsgi.readthedocs.io/en/latest/Async.html#suspend-resume

    :type: bool
    """


def suspend():
    """Suspends handling of current coroutine/green thread and passes control
    to the next async core.

    * http://uwsgi.readthedocs.io/en/latest/Async.html#suspend-resume

    :type: bool
    """


def i_am_the_lord(legion_name):
    """Returns flag indicating whether you are the lord
    of the given legion.

    * http://uwsgi.readthedocs.io/en/latest/Legion.html#legion-api

    :param str|unicode legion_name:

    :rtype: bool
    """


def i_am_the_spooler():
    """Returns flag indicating whether you are the Spooler.

    :rtype: bool
    """


def in_farm(name):
    """Returns flag indicating whether you are (mule) belong
    to the given farm.

    :param str|unicode name: Farm name.

    :rtype: bool
    """


def is_a_reload():
    """Returns flag indicating whether reloading mechanics is used.

    :rtype: bool
    """


def is_connected(fd):
    """Checks the given file descriptor.

    :param int fd: File descriptor

    :rtype: bool
    """


def is_locked(lock_num=0):
    """Checks for the given lock.

    .. note:: Lock 0 is always available.

    :param int lock_num: Lock number.

    :rtype: bool

    :raises ValueError: For Spooler or invalid lock number
    """


def listen_queue(socket_num):
    """Returns listen queue (backlog size) of the given socket.

    :param int socket_num: Socket number.

    :rtype: bool

    :raises ValueError: If socket is not found
    """


def lock(lock_num=0):
    """Sets the given lock.

    .. note:: Lock 0 is always available.

    :param int lock_num: Lock number.

    :rtype: None

    :raises ValueError: For Spooler or invalid lock number
    """


def log(message):
    """Logs a message.

    :param str|unicode message:

    :rtype: bool
    """


def log_this_request():
    """Instructs uWSGI to log current request data.

    :rtype: None
    """


def logsize():
    """Returns current log size.

    :rtype|long
    """


def loop():
    """Returns current event loop name or None if loop is not set.

    :rtype: st|unicode|None
    """


def lord_scroll(legion_name):
    """Returns a Lord scroll for the Legion.

    * http://uwsgi.readthedocs.io/en/latest/Legion.html#lord-scroll-coming-soon

    :param str|unicode legion_name:

    :rtype: bool
    """


def masterpid():
    """Return the process identifier (PID) of the uWSGI master process.

    :rtype: int
    """


def mem():
    """Returns memory usage tuple of ints: (rss, vsz).

    :rtype: tuple
    """


def metric_dec(key, value=1):
    """Decrements the specified metric key value by the specified value.

    :param str|unicode key:

    :param int value:

    :rtype: bool
    """


def metric_div(key, value=1):
    """Divides the specified metric key value by the specified value.

    :param str|unicode key:

    :param int value:

    :rtype: bool
    """


def metric_inc(key, value=1):
    """Increments the specified metric key value by the specified value.

    :param str|unicode key:

    :param int value:

    :rtype: bool
    """


def metric_mul(key, value=1):
    """Multiplies the specified metric key value by the specified value.

    :param str|unicode key:

    :param int value:

    :rtype: bool
    """


def metric_get(key):
    """Returns metric value by key.

    :param str|unicode key:

    :rtype: int|long
    """


def metric_set(key, value):
    """Sets metric value.

    :param str|unicode key:

    :param int|long value:

    :rtype: bool
    """


def metric_set_max(key, value):
    """Sets metric value if it is greater that the current one.

    :param str|unicode key:

    :param int|long value:

    :rtype: bool
    """


def metric_set_min(key, value):
    """Sets metric value if it is less that the current one.

    :param str|unicode key:

    :param int|long value:

    :rtype: bool
    """


def micros():
    """Returns uWSGI clock microseconds.

    :rtype|long
    """


def mule_get_msg(signals=None, farms=None, buffer_size=65536, timeout=-1):
    """Block until a mule message is received and return it.

    This can be called from multiple threads in the same programmed mule.

    .. warning:: Bytes are returned for Python 3.

    :param bool signals: Whether to manage signals.

    :param bool farms: Whether to manage farms.

    :param int buffer_size:

    :param int timeout: Seconds.

    :rtype: bytes|str

    :raises ValueError: If not in a mule.
    """


def mule_id():
    """Returns current mule ID.

    :rtype: int
    """


def mule_msg(message, mule_farm=None):
    """Sends a message to a mule(s)/farm.

    :param str|unicode message:

    :param mule_farm: Mule ID, or farm name.

    :rtype: bool

    :raises ValueError: If no mules, or mule ID or farm name is not recognized.
    """


def offload(filename):
    """Offloads a file.

    .. warning:: Currently not implemented.

    :param filename:

    :rtype: bytes|str

    :raises ValueError: If unable to offload.
    """


def parsefile(fpath):
    """Parses the given file.

    Currently implemented only Spooler file parsing.

    :param str|unicode fpath:

    :rtype: None
    """


def ready():
    """Returns flag indicating whether we are ready to handle requests.

    :rtype: bool
    """


def ready_fd():
    """Returns flag indicating whether file description related to request is ready.

    :rtype: bool
    """


def recv(fd, maxsize=4096):
    """Reads data from the given file descriptor.

    .. warning:: Bytes are returned for Python 3.

    :param int fd:

    :param int maxsize: Chunk size (bytes).

    :rtype: bytes|str
    """


def register_rpc(name, func):
    """Registers RPC function.

    * http://uwsgi.readthedocs.io/en/latest/RPC.html

    :param str|unicode name:

    :param callable func:

    :rtype: bool

    :raises ValueError: If unable to register function
    """


def register_signal(number, target, func):
    """Registers a signal handler.

    :param int number: Signal number.

    :param str|unicode target:

        * ``workers``  - run the signal handler on all the workers
        * ``workerN`` - run the signal handler only on worker N
        * ``worker/worker0`` - run the signal handler on the first available worker
        * ``active-workers`` - run the signal handlers on all the active [non-cheaped] workers

        * ``mules`` - run the signal handler on all of the mules
        * ``muleN`` - run the signal handler on mule N
        * ``mule/mule0`` - run the signal handler on the first available mule

        * ``spooler`` - run the signal on the first available spooler
        * ``farmN/farm_XXX``  - run the signal handler in the mule farm N or named XXX

        * http://uwsgi.readthedocs.io/en/latest/Signals.html#signals-targets

    :param callable func:

    :rtype: None

    :raises ValueError: If unable to register
    """


def reload():
    """Gracefully reloads uWSGI.

    * http://uwsgi.readthedocs.io/en/latest/Management.html#reloading-the-server

    :rtype: bool
    """


def request_id():
    """Returns current request number (handled by worker on core).

    :rtype: int
    """


def route(name, args_str):
    """Registers a named route for internal routing subsystem.

    :param str name: Route name

    :param args_str: Comma-separated arguments string.

    :rtype: int
    """


def rpc(address, func_name, *args):
    """Performs an RPC function call with the given arguments.

    * http://uwsgi.readthedocs.io/en/latest/RPC.html

    .. warning:: Bytes are returned for Python 3.

    :param str|unicode address:

    :param str|unicode func_name: Function name to call.

    :param args:

    :rtype: bytes|str

    :raises ValueError: If unable to RPC function.
    """


def rpc_list():
    """Returns registered RPC functions names.

    :rtype: tuple
    """


def scrolls(legion_name):
    """Returns a list of Legion scrolls defined on cluster.

    :param str|unicode legion_name:

    :rtype: list
    """


def send(df_or_data, data=None):
    """Puts data into file descriptor.

    * One argument. Data to write into request file descriptor.
    * Two arguments. 1. File descriptor; 2. Data to write.

    :param df_or_data:

    :param data:

    :rtype: bool
    """


def sendfile(fd_or_name, chunk_size=0, start_pos=0, filesize=0):
    """Runs a sendfile.

    :param int|str|unicode fd_or_name: File path or descriptor number.

    :param int chunk_size: Not used.

    :param int start_pos:

    :param int filesize: Filesize. If ``0`` will be determined automatically.

    :rtype: bool|None
    """


def set_logvar(name, value):
    """Sets log variable.

    :param str|unicode name:

    :param str|unicode value:

    :rtype: None
    """


def set_user_harakiri(timeout=0):
    """Sets user level harakiri.

    :param int timeout: Seconds. ``0`` disable timer.

    :rtype: None
    """


def set_warning_message(message):
    """Sets a warning. This will be reported by pingers.

    :param str|unicode message:

    :rtype: bool
    """


def setprocname(name):
    """Sets current process name.

    :param str|unicode name:

    :rtype: bool
    """


def signal(num, remote=None):
    """Sends the signal to master or remote.

    :param num: Signal number.

    :param str|unicode remote: Remote address.

    :rtype: None

    :raises ValueError: If remote rejected the signal.

    :raises IOError: If unable to deliver to remote.
    """


def signal_received():
    """Get the number of the last signal received.

    Used in conjunction with ``signal_wait``.

    * http://uwsgi-docs.readthedocs.io/en/latest/Signals.html#signal-wait-and-signal-received

    :rtype: int
    """


def signal_registered(num):
    """Verifies the given signal has been registered.

    :param int num:

    :rtype: bool|None
    """


def signal_wait(num=None):
    """Waits for the given of any signal.

    Block the process/thread/async core until a signal is received. Use signal_received to get the number of
    the signal received. If a registered handler handles a signal, signal_wait will be interrupted and the actual
    handler will handle the signal.

    * http://uwsgi-docs.readthedocs.io/en/latest/Signals.html#signal-wait-and-signal-received

    :param int num:

    :rtype: str|unicode

    :raises SystemError: If something went wrong.
    """


def sockets():
    """Returns a current list file descriptors for registered sockets.

    :rtype: list[int]
    """


def stop():
    """Stops uWSGI.

    :rtype: bool|None
    """


def total_requests():
    """Returns the total number of requests managed so far by the pool of uWSGI workers.

    :rtype: int
    """


def unlock(lock_num=0):
    """Unlocks the given lock.

    .. note:: Lock 0 is always available.

    :param int lock_num: Lock number.

    :rtype: None

    :raises ValueError: For Spooler or invalid lock number
    """


def wait_fd_read(fd, timeout=None):
    """Suspends handling of the current request until there is something
    to be read on file descriptor.

    May be called several times before yielding/suspending
    to add more file descriptors to the set to be watched.

    * http://uwsgi-docs.readthedocs.io/en/latest/Async.html#waiting-for-i-o

    :param int fd: File descriptor number.

    :param int timeout: Timeout. Default:  infinite.

    :rtype: bytes|str

    :raises IOError: If unable to read.
    """


def wait_fd_write(fd, timeout=None):
    """Suspends handling of the current request until there is nothing more
    to be written on file descriptor.

    May be called several times to add more file descriptors to the set to be watched.

    * http://uwsgi-docs.readthedocs.io/en/latest/Async.html#waiting-for-i-o

    :param int fd: File descriptor number.

    :param int timeout: Timeout. Default:  infinite.

    :rtype: bytes|str

    :raises IOError: If unable to read.
    """


def websocket_handshake(security_key=None, origin=None, proto=None):
    """Waits for websocket handshake.

    :param str|unicode security_key: Websocket security key to use.

    :param str|unicode origin: Override ``Sec-WebSocket-Origin``.

    :param str|unicode proto: Override ``Sec-WebSocket-Protocol``.

    :rtype: None

    :raises IOError: If unable to complete handshake.
    """


def websocket_recv(request_context=None):
    """Receives data from websocket.

    :param request_context:

    :rtype: bytes|str

    :raises IOError: If unable to receive a message.
    """


def websocket_recv_nb(request_context=None):
    """Receives data from websocket (non-blocking variant).

    :param request_context:

    :rtype: bytes|str

    :raises IOError: If unable to receive a message.
    """


def websocket_send(message, request_context=None):
    """Sends a message to websocket.

    :param str message: data to send

    :param request_context:

    :raises IOError: If unable to send a message.
    """


def websocket_send_binary(message, request_context=None):
    """Sends binary message to websocket.

    :param str message: data to send

    :param request_context:

    :raises IOError: If unable to send a message.
    """


def worker_id():
    """Returns current worker ID.

    :rtype: int
    """


def workers():
    """Gets statistics for all the workers for the current server.

    Returns tuple of dicts.

    :rtype: tuple[dict]
    """



from ..base import ParametrizedValue


class Encoder(ParametrizedValue):

    name_separator = ' '


class EncoderPrefix(Encoder):
    """Add a raw prefix to each log msg."""

    name = 'prefix'

    def __init__(self, value):
        """

        :param str|unicode value: Value to be used as affix

        """
        super(EncoderPrefix, self).__init__(value)


class EncoderSuffix(EncoderPrefix):
    """Add a raw suffix to each log msg"""

    name = 'suffix'


class EncoderNewline(Encoder):
    """Add a newline char to each log msg."""

    name = 'nl'


class EncoderGzip(Encoder):
    """Compress each msg with gzip (requires zlib)."""

    name = 'gzip'


class EncoderCompress(Encoder):
    """Compress each msg with zlib compress (requires zlib)."""

    name = 'compress'


class EncoderFormat(Encoder):
    """Apply the specified format to each log msg."""

    name = 'format'

    def __init__(self, template):
        """

        :param str|unicode template: Template string.
            Available variables are listed in ``FormatEncoder.Vars``.

        """
        super(EncoderFormat, self).__init__(template)

    class vars(object):
        """Variables available to use."""

        MESSAGE = '${msg}'
        """Raw log message (newline stripped)."""

        MESSAGE_NEWLINE = '${msgnl}'
        """Raw log message (with newline)."""

        TIME = '${unix}'
        """Current unix time."""

        TIME_MS = '${micros}'
        """Current unix time in microseconds."""

        # todo consider adding ${strftime:xxx} - strftime using the xxx format


class EncoderJson(EncoderFormat):
    """Apply the specified format to each log msg with each variable json escaped."""

    name = 'json'
